# -*- coding: iso-8859-1 -*-
"""
This module contains the specific optimisation algorithms for TopasOpt. Most functionality is defined in TopasOptBaseClass,
which other optimisers inherit from.
"""
import subprocess
import jsonpickle
from matplotlib import pyplot as plt
# matplotlib.use('Agg')  # if having trouble with generating figures through ssh, this resolves...
import shutil
from scipy.optimize import minimize
from scipy import stats
from pathlib import Path
import numpy as np
import sys, os
from bayes_opt import BayesianOptimization
from bayes_opt import UtilityFunction
from bayes_opt.logger import JSONLogger
from bayes_opt.util import load_logs
from bayes_opt.event import Events
from sklearn.gaussian_process.kernels import Matern
import logging
from .utilities import bcolors, FigureSpecs, newJSONLogger, ReadInLogFile, PlotLogFile
import stat
from scipy.optimize import rosen

ch = logging.StreamHandler()
formatter = logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.INFO)  # This toggles all the logging in your app
logger.propagate = False


def _import_from_absolute_path(fullpath, global_name=None):
    """
    Dynamic script import using full path.
    This is required to enable mapping to the location of the script generation function and the objective funciton,
    which are not known in advance.
    `credit here`_<https://stackoverflow.com/questions/3137731/is-this-correct-way-to-import-python-scripts-residing-in-arbitrary-folders>
    """

    script_dir, filename = os.path.split(fullpath)
    script, ext = os.path.splitext(filename)

    sys.path.insert(0, script_dir)
    try:
        module = __import__(script)
        if global_name is None:
            global_name = script
        globals()[global_name] = module
        sys.modules[global_name] = module
    finally:
        del sys.path[0]


class TopasOptBaseClass:
    """
    There are many overlapping functionalities required by all optimisation algorithms: logging, calculation of objective function,
    generation of models...etc. All of these common methods are contained this base class which other optimisation methods inherit.
    This class is not intended to be used in isolation and won't work if you try.
    An important thing to note is that the variables from all optimisers are defined here, which means some of these options
    are specific to a particular optimiser. We may find a more elegant way to handle this in future!

    :param optimisation_params: Parameters to be optimised.
    :type optimisation_params: list or array
    :param BaseDirectory: Place where all the topas simulation results are stored
    :type BaseDirectory: string
    :param SimulationName: Specific folder for this simulation
    :type SimulationName: string
    :param OptimisationDirectory: location that TopasObjectiveFunction and GenerateTopasScript are located
    :type OptimisationDirectory: string or Path
    :param ReadMeText: If supplied, is written to a readme file in BaseDirectory
    :type ReadMeText: string, optional
    :param TopasLocation: location of topas installation. Default is ~/topas if you follow the topas instructions.
    :type TopasLocation: string or pathlib.Path, optional
    :param ShellScriptHeader: Header to place at the start of the bash file that runs the topas model. This header
        should contain all the commands you normally perform to set up your terminal environment.
    :type ShellScriptHeader: string, optional
    :param Overwrite: if True, will automatically overwrite any existing files. If False, will ask first (safer)
    :type Overwrite: bool, optional
    :param KeepAllResults: if True, all results kept, if false, only most recent results are kept. Note that in either
        case the log files contain the info from all cases, it's just a matter of whether you want to store every iteration
        which can take a lot of space
    :type KeepAllResults: bool, optional
    """

    def __init__(self, optimisation_params, BaseDirectory, SimulationName, OptimisationDirectory,
                 ReadMeText=None,
                 TopasLocation='~/topas/',
                 ShellScriptHeader=None, Overwrite=False, KeepAllResults=True):
        """
        init method for all optimisers. input options are in class docstring
        """

        if 'TopasOptBaseClass' in str(self.__class__):
            logger.error('TopasOptBaseClass should not be called directly; it only exists for other optimisers'
                         'to inherit. Quitting')
            sys.exit(1)

        optimisation_params = self._convert_optimisation_params_to_numpy(optimisation_params)
        self.ReadMeText = ReadMeText  # this gets written to base directory
        self.ShellScriptHeader = ShellScriptHeader
        self.KeepAllResults = KeepAllResults
        # attempt the absolute imports from the optimisation directory:
        self.BaseDirectory = BaseDirectory
        self.OptimisationDirectory = OptimisationDirectory

        if not os.path.isdir(BaseDirectory):
            logger.error(
                f'{bcolors.FAIL}Input BaseDirectory "{BaseDirectory}" does not exist. Exiting. {bcolors.ENDC}')
            sys.exit()
        self.SimulationName = SimulationName
        _LogFileLoc  = Path(self.BaseDirectory) / self.SimulationName
        _LogFileLoc  = _LogFileLoc  / 'logs'
        self._LogFileLoc = str(_LogFileLoc  / 'OptimisationLogs.txt')
        self.Itteration = 0
        self.ItterationStart = 0
        self._optimisation_params = optimisation_params
        # the starting values of our optimisation parameters are defined from the default geometry
        self.ParameterNames = optimisation_params['ParameterNames']
        self.StartingValues = optimisation_params['start_point'].astype(float)
        if self.StartingValues is None:
            logger.error('you must define a start point')
            sys.exit()
        else:
            self.x = self.StartingValues
        self.UpperBounds = optimisation_params['UpperBounds']
        self.LowerBounds = optimisation_params['LowerBounds']
        self.MaxItterations = int(optimisation_params['Nitterations'])
        self._CreateVariableDictionary([self.StartingValues])
        self.SuggestionsProbed = 0  # always starts at 0
        self.Overwrite = Overwrite
        self.AllObjectiveFunctionValues = []

        if '~' in TopasLocation:
            TopasLocation = os.path.expanduser(TopasLocation)
        self.TopasLocation = Path(TopasLocation)
        self._testing_mode = False  # overwrite if True
        if 'testing_mode' in str(self.TopasLocation):
            self._testing_mode = True
            logger.warning(f'Entering test mode because topas location = {self.TopasLocation}')


        # this is where we try to load the user defined model generator and objective function
        try:
            _import_from_absolute_path(Path(self.OptimisationDirectory) / 'GenerateTopasScripts.py')
        except ModuleNotFoundError as e:
            logger.error(f'Failed to import required file at {str(Path(self.OptimisationDirectory) / "GenerateTopasScript.py")}.'
                         f'\nQuitting')
            raise e
        try:
            _import_from_absolute_path(Path(self.OptimisationDirectory) / 'TopasObjectiveFunction.py')
        except ModuleNotFoundError as e:
            logger.error(f'Failed to import required file at {str(Path(self.OptimisationDirectory) / "TopasObjectiveFunction.py")}.'
                         f'\nQuitting')
            raise e
        self.TopasScriptGenerator = GenerateTopasScripts.GenerateTopasScripts
        self.TopasObjectiveFunction = TopasObjectiveFunction.TopasObjectiveFunction
        self._CheckInputData()

    def _convert_optimisation_params_to_numpy(self, optimisation_params):
        """
        This simply checks if any of the supplied parameters are a list instead of an array, and converts them
        if so. will also cast all parameters to floats - if user enntered e.g. 1, python treats it as int
        """
        for param_key in list(optimisation_params.keys()):
            if param_key == 'ParameterNames' or param_key == 'Nitterations':
                continue
            if isinstance(optimisation_params[param_key], list):
                optimisation_params[param_key] = np.array(optimisation_params[param_key])
            if not isinstance(optimisation_params[param_key], np.ndarray):
                logger.error(f'{bcolors.FAIL} optimisation param {param_key} should be a list or an array....')
            optimisation_params[param_key].astype(float)

        return optimisation_params

    def _CreateVariableDictionary(self, x):
        """
        Use the input information to create a dictionary of geometric inputs.
        This is what gets passed to build the topas model
        x is the current list of parameter guesses.
        """

        if np.ndim(x) == 1:
            self.VariableDict = {self.ParameterNames[i]: x[i] for i in range(len(self.ParameterNames))}
        elif np.ndim(x) == 2:
            self.VariableDict = {self.ParameterNames[i]: x[0][i] for i in range(len(self.ParameterNames))}
        else:
            logger.error('seomthing wrong with input parameter format...quitting')
            sys.exit(1)

        for key in self.VariableDict.keys():
            # this just kept happening so put this ugle brutfe forxe method in
            if type(self.VariableDict[key]) is np.ndarray:
                self.VariableDict[key] = self.VariableDict[key][0]
        # finally add in the beamlet size:

    def _EmptySimulationFolder(self):
        """
        If there is already stuff in the simulation folder, ask for user permission to empty and continue
        """
        SimName = str(Path(self.BaseDirectory) / self.SimulationName)
        if os.listdir(SimName) and (not self.Overwrite):
            logger.warning(f'Directory {SimName} is not empty; if you continue it will be emptied.'
                           f'\nType y to continue.'
                           f'\nYou can set Overwrite=True to disable this warning')

        if self.Overwrite:
            UserOverwrite = 'y'
        else:
            UserOverwrite = input()

        if (UserOverwrite.lower() == 'n') or (UserOverwrite.lower() == 'no'):
            logger.warning('Not overwriting and quitting')
            sys.exit(1)
        elif (UserOverwrite.lower() == 'y') or (UserOverwrite.lower() == 'yes'):
            logger.warning('emptying simulation folder')
            for filename in os.listdir(SimName):
                file_path = os.path.join(SimName, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except Exception as e:
                    logger.error(f'Failed to delete {file_path}.')
                    raise e

    def _CheckInputData(self):
        """
        check that the user has put in reasonable parameters:

        - Checks if the number of parameters in ParameterNames, StartingValues, UpperBounds, LowerBounds match
        - Checks that StartingValues is actually within the provided bounds
        """

        # do the number of parameters match?
        if not np.size(self.ParameterNames) == np.size(self.StartingValues):
            logger.error(f'{bcolors.FAIL} size of ParameterNames does not match size of StartingValues{bcolors.ENDC}')
            sys.exit(1)
        if not np.size(self.StartingValues) == np.size(self.UpperBounds):
            print(f'{bcolors.FAIL} size of StartingValues does not match size of UpperBounds{bcolors.ENDC}')
            sys.exit(1)
        if not np.size(self.UpperBounds) == np.size(self.LowerBounds):
            print(f'{bcolors.FAIL} size of UpperBounds does not match size of LowerBounds{bcolors.ENDC}')
            sys.exit(1)

        for i, Paramter in enumerate(self.ParameterNames):
            try:
                if self.StartingValues[0][i] < self.LowerBounds[i]:
                    print(f'{bcolors.FAIL}For {Paramter}, Starting value {self.StartingValues[0][i]} is less than '
                          f'Lower bound {self.LowerBounds[i]}{bcolors.ENDC}')
                    sys.exit(1)
                elif self.StartingValues[0][i] > self.UpperBounds[i]:
                    print(f'{bcolors.FAIL}For {Paramter}, Starting value {self.StartingValues[0][i]} is greater '
                          f'than upper bound {self.UpperBounds[i]}{bcolors.ENDC}')
                    sys.exit(1)
            except IndexError:
                if self.StartingValues[i] < self.LowerBounds[i]:
                    print(f'{bcolors.FAIL}For {Paramter}, Starting value {self.StartingValues[i]} is less than '
                          f'Lower bound {self.LowerBounds[i]}{bcolors.ENDC}')
                    sys.exit(1)
                elif self.StartingValues[i] > self.UpperBounds[i]:
                    print(f'{bcolors.FAIL}For {Paramter}, Starting value {self.StartingValues[i]} is greater '
                          f'than upper bound {self.UpperBounds[i]}{bcolors.ENDC}')
                    sys.exit(1)

        self._CreateVariableDictionary(self.StartingValues)

        # make sure topas binary exists
        if not self._testing_mode:
            if not os.path.isfile(self.TopasLocation / 'bin' / 'topas'):
                logger.error(f'{bcolors.FAIL}could not find topas binary at \n{self.TopasLocation}'
                             f'\nPlease initialise with TopasLocation pointing to the topas installation location.'
                             f'\nQuitting{bcolors.ENDC}')
                sys.exit(1)

    def _GenerateTopasModel(self, x):
        """
        Generates a topas model with the latest parameters as well as a shell script called RunAllFiles.sh to run it.
        """

        self.TopasScripts, self.TopasScriptNames = self.TopasScriptGenerator(self.BaseDirectory, self.Itteration, **self.VariableDict)
        self.ScriptsToRun = []
        for i, script_name in enumerate(self.TopasScriptNames):
            script_name = script_name + '_itt_' + str(self.Itteration) + '.tps'
            self.ScriptsToRun.append(script_name)
            f = open(str(Path(self.BaseDirectory) / self.SimulationName / 'TopasScripts' / script_name), 'w')
            for line in self.TopasScripts[i]:
                f.write(line)
                f.write('\n')

        self._GenerateRunIterationShellScript()

    def _setup_topas_emulator(self):
        """
        despite it's fancy sounding name, this isn't really an emulator at all
        What it does is create a bash script that literally does nothing except print hello
        But this allows us to test most of the functionality of this code when we are in testing mode,
        without having to run time consuming topas simulations
        """
        if not os.path.isdir(Path(self.BaseDirectory) / self.SimulationName / 'bin'):
            os.mkdir(Path(self.BaseDirectory) / self.SimulationName / 'bin')

        EmulatorLocation = Path(self.BaseDirectory) / self.SimulationName / 'bin' / 'topas'
        if os.path.isfile(EmulatorLocation):
            # I don't think I should need to do this; the file should be overwritten if it exists, but this doesn't
            # seem to be working so deleting it.
            os.remove(EmulatorLocation)
        self.TopasLocation = EmulatorLocation.parent.parent

        f = open(EmulatorLocation, 'w+')
        f.write('echo "Hello from topas emulator! I dont do anything except print this"')
        # change file permissions:
        st = os.stat(EmulatorLocation)
        os.chmod(EmulatorLocation, st.st_mode | stat.S_IEXEC)
        f.close()

    def _GenerateRunIterationShellScript(self):
        """
        This will generate a bash script called 'RunAllFiles', which, funnily enough, can be used to run all files generated!
        """
        ShellScriptLocation = str(Path(self.BaseDirectory) / self.SimulationName / 'TopasScripts' / 'RunIteration.sh')
        if os.path.isfile(ShellScriptLocation):
            # I don't think I should need to do this; the file should be overwritten if it exists, but this doesn't
            # seem to be working so deleting it.
            os.remove(ShellScriptLocation)
        self.ShellScriptLocation = ShellScriptLocation
        f = open(ShellScriptLocation, 'w+')
        # set up the environment etc.
        if self.ShellScriptHeader is None:
            f.write('# !/bin/bash')
            f.write('\n\n# This script sets up the topas environment then runs all listed files\n\n')
            f.write('\nexport TOPAS_G4_DATA_DIR=~/G4Data\n')
        else:
            f.write(self.ShellScriptHeader)

        # add in all topas scripts which need to be run:
        for script_name in self.ScriptsToRun:
            file_loc = str(Path(self.BaseDirectory) / self.SimulationName / 'TopasScripts' / script_name)
            f.write('echo "Beginning analysis of: ' + script_name + '"')
            f.write('\n')
            f.write('(time ' + str(self.TopasLocation) + '/bin/topas ' + script_name + ') &> ../logs/TopasLogs/' + script_name)
            f.write('\n')
        # change file permissions:
        st = os.stat(ShellScriptLocation)
        os.chmod(ShellScriptLocation, st.st_mode | stat.S_IEXEC)
        f.close()

    def _RunTopasModel(self):
        """
        This invokes a bash subprocess to run the current model
        """
        print(f'{bcolors.OKBLUE}Topas: Running file: \n{self.ShellScriptLocation}')
        ShellScriptPath = str(Path(self.BaseDirectory) / self.SimulationName / 'TopasScripts')
        cmd = subprocess.run(['bash', self.ShellScriptLocation], cwd=ShellScriptPath)
        if cmd.returncode == 0:
            print(f'{bcolors.OKBLUE}Analysis complete{bcolors.ENDC}')
        else:
            logger.error(f'RunIteration.sh failed with exit code {cmd.returncode}.'
                         f'\nSuggestion: look at {Path(self.OptimisationDirectory) / self.SimulationName / "Logs" / "TopasLogs"} '
                         f'\nto figure out what went wrong...'
                         f' Quitting')
            sys.exit(1)

    def _UpdateOptimisationLogs(self, x, OF):
        """
        Just a simple function to keep track of the objective function in the logs folder

        :param x: the current parameter values
        :type x: array
        :param OF: the current objective function value
        """

        with open(self._LogFileLoc , 'a') as f:
            Entry = f'Itteration: {self.Itteration}'
            for i, Parameter in enumerate(self.ParameterNames):
                try:
                    Entry = Entry + f', {Parameter}: {x[0][i]: 1.2f}'
                except IndexError:
                    Entry = Entry + f', {Parameter}: {x[i]: 1.2f}'

            try:
                Entry = Entry + f', _target_prediction_mean: {self._target_prediction_mean[-1]: 1.2f}'
                Entry = Entry + f', _target_prediction_std: {self._target_prediction_std[-1]: 1.2f}'
            except AttributeError:
                # these parameters are only available for bayes optimisation
                pass
            except IndexError:
                # for the first entry
                Entry = Entry + f', _target_prediction_mean: NaN'
                Entry = Entry + f', _target_prediction_std: NaN'

            Entry = Entry + f', ObjectiveFunction: {OF: 1.2f}\n'
            f.write(Entry)
        print(f'{bcolors.OKGREEN}{Entry}{bcolors.ENDC}')

    def _write_final_log_entry(self):
        """
        This method can optionally be called when an optimiser has finished running.
        It reads in the logs, then prints a message at the end summarising the best found solution.
        """
        ResultsDict = ReadInLogFile(self._LogFileLoc)

        best_iteration = np.argmin(ResultsDict['ObjectiveFunction'])
        best_OF = ResultsDict['ObjectiveFunction'][best_iteration]
        ResultsDict.pop('ObjectiveFunction')
        ResultsDict.pop('Itteration')
        try:
            ResultsDict.pop('_target_prediction_mean')
            ResultsDict.pop('_target_prediction_std')
        except KeyError:
            pass
        # what's left is the results (hopefully no one starts updating the log format!)
        ParameterValues = list(ResultsDict.values())
        best_params = np.array([param_list[best_iteration] for param_list in ParameterValues])
        sort_ind = np.argsort(self.ParameterNames)  # need to sort parameters alphabetically
        best_params = best_params[sort_ind]

        with open(self._LogFileLoc , 'a') as f:
            Entry = f'\nBest parameter set: '

            Entry = Entry + f'Itteration: {best_iteration}.'
            for i, Parameter in enumerate(sorted(self.ParameterNames)):
                Entry = Entry + f', {Parameter}: {best_params[i]: 1.2f}'

            Entry = Entry + f', ObjectiveFunction: {best_OF: 1.2f}\n'
            f.write(Entry)
        print(f'{bcolors.OKGREEN}{Entry}{bcolors.ENDC}')

    def _Plot_Convergence(self):
        """
        just generates the current convergence plot from the log file using the function in utilities
        """

        SaveLoc = Path(self.BaseDirectory) / self.SimulationName
        SaveLoc = SaveLoc / 'logs' / 'ConvergencePlot.png'
        PlotLogFile(self._LogFileLoc, SaveLoc)

    def _CopySelf(self):
        """
        Copies all class attributes to a a (human readable) json file called 'SimulationSettings'.
        This is so you can exactly which settings were used to generate a given simulation
        """

        Filename = Path(self.BaseDirectory) / Path(self.SimulationName) / 'OptimisationSettings.json'

        Attributes = jsonpickle.encode(self, unpicklable=True, max_depth=4)
        f = open(str(Filename), 'w')
        f.write(Attributes)

    def _ConvertDictToVariables(self, x_new):
        """
        I'm trying to keep as much of the underlying code base compatible with multiple methods, so convert the x_new
        the bayes code gives me into what the rest of my code expects

        Note the bayes code sends in the dict in alphabetical order; this methods also corrects that such that the initial
        order of variables is maintained
        """
        if not isinstance(x_new, dict):
            # then no conversion needed
            self.x = x_new
            return
        x = []
        for i, paramNames in enumerate(x_new):
            x.append(x_new[self.ParameterNames[i]])
        self.x = np.array(x, ndmin=2)

    def _empty_results_folder(self):
        """
        if KeepAllResults is false, this function is called which removes all existing files
        in the results folder. It is called just before new scripts are run such that the latest results
        will be kept
        """
        ResultsLocation = str(Path(self.BaseDirectory) / self.SimulationName / 'Results')
        for filename in os.listdir(ResultsLocation):
            file_path = os.path.join(ResultsLocation, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                logger.warning(f'Failed to delete {file_path} from results folder. Reason: {e}. continuing...')

    # public methods
    
    def BlackBoxFunction(self, x_new):
        """
        Called Black Box function in the spirit of bayesian optimisation, this function simply takes the most recent
        parameter guesses, and solves the model.
        """

        if self._testing_mode:
            # this is a special section only intended for development, unit testing, etc.
            # if you are here, it means 'testing_mode' is within your TopasLocation
            self._ConvertDictToVariables(x_new)
            self._CreateVariableDictionary(self.x)
            self._GenerateTopasModel(self.x)
            if not self.KeepAllResults:
                self._empty_results_folder()
            self._RunTopasModel()
            self.OF = self.TopasObjectiveFunction(Path(self.BaseDirectory) / self.SimulationName / 'Results',
                                                  self.Itteration)
            # this is the part where we would normally read in the results and assess the objective function, but
            # since we use TopasEmulator, there are no results. therefore we need to 'insert' them.
            if self.x.shape[0] == 2:
                x = self.x
            elif self.x.shape[0] == 1:
                x = self.x[0]
            else:
                logger.error('how have you managed this.')
                sys.exit(1)
            self.OF = np.min([rosen(x), 100])  # rosenbrock function can get huge, so just cap it at 100.
            self.AllObjectiveFunctionValues.append(self.OF)
            self._UpdateOptimisationLogs(self.x, self.OF)
            self._Plot_Convergence()
            self.Itteration = self.Itteration + 1
        else:
            # this is the normal optimisation
            self._ConvertDictToVariables(x_new)
            self._CreateVariableDictionary(self.x)
            self._GenerateTopasModel(self.x)
            if not self.KeepAllResults:
                self._empty_results_folder()
            self._RunTopasModel()
            self.OF = self.TopasObjectiveFunction(Path(self.BaseDirectory) / self.SimulationName / 'Results', self.Itteration)
            self.AllObjectiveFunctionValues.append(self.OF)
            self._UpdateOptimisationLogs(self.x, self.OF)
            self._Plot_Convergence()
            self.Itteration = self.Itteration + 1

        if 'BayesianOptimiser' in str(self.__class__):
            # bit of a hack since bayesian optimise will seek maximum
            return -self.OF
        else:
            return self.OF

    def SetUpDirectoryStructure(self):
        """
        Method to set up directory structure. This will attempt to empty the directory if it already exists.
        If Overwrite=False, it will ask first, otherwise just do it.
        Also writes the readme text if that exists, and copies all attributes of self to a json file.
        """

        FullSimName = Path(self.BaseDirectory) / self.SimulationName
        if not os.path.isdir(FullSimName):
            os.mkdir(FullSimName)
        self._EmptySimulationFolder()
        self._CopySelf()
        os.mkdir(Path(FullSimName) / 'logs')
        os.mkdir(Path(FullSimName) / 'logs' / 'TopasLogs')
        if 'BayesianOptimiser' in str(self.__class__):
            os.mkdir(Path(FullSimName) / 'logs' / 'SingleParameterPlots')
        os.mkdir(Path(FullSimName) / 'TopasScripts')
        os.mkdir(Path(FullSimName) / 'Results')

        if self.ReadMeText:
            f = open(FullSimName / 'readme.txt','w')
            f.write(self.ReadMeText)

        if self._testing_mode:
            self._setup_topas_emulator()


class NelderMeadOptimiser(TopasOptBaseClass):
    """
    Implementation of Nelder-Mead based on scipy. Other options are defined in TopasOptBaseClass

    :param NM_StartingSimplex: This is a Nelder-Mead specific parameter which controls the size of the
        starting simplex. A value of .1 will create a starting simplex that is spans 10% of the starting values,
        which is the default behavior. Alternatively one can specify the starting simplex, e.g. for a 2D function
        starting_sim = [[0.9, 0.9], [0.72, 0.9], [0.9, 0.72]]
    :type NM_StartingSimplex: None or float or array-like, optional
    """

    def __init__(self, NM_StartingSimplex=None, **kwds):
        """
        init function for NelderMeadOptimiser
        """
        self.NM_StartingSimplex = NM_StartingSimplex
        super().__init__(**kwds)
        if self.NM_StartingSimplex:  # nb None evaluates as False
            self.StartingSimplexSupplied = True
            self._GenerateStartingSimplex()

    def _GenerateStartingSimplex(self):
        """
        Enbles the user to enter their own starting simplex, or optionally scale the starting simplex by some fraction.
        This is copied from the scipy source code (optimize around line 690), with a variable version of nonzdelt
        The default is 0.1
        """

        if isinstance(self.NM_StartingSimplex, float):
            # then we generate a simplex based on the starting values
            N = len(self.StartingValues)
            nonzdelt = self.NM_StartingSimplex
            zdelt = self.NM_StartingSimplex / 200  # based on the scipy code where zdelt - nonzdelt/200
            sim = np.empty((N + 1, N), dtype=self.StartingValues.dtype)
            sim[0] = self.StartingValues
            for k in range(N):
                y = np.array(self.StartingValues, copy=True)
                if y[k] != 0:
                    store_yk_temp = y[k]  # yuck, surely someone can do better than this
                    y[k] = (1 + nonzdelt) * y[k]
                    if y[k] > self.UpperBounds[k]:
                        y[k] = (1 - nonzdelt) * store_yk_temp
                    if y[k]  < self.LowerBounds[k]:
                        logger.warning('unable to generate starting simplex within bounds. Simplex will be clipped to bounds.'
                                       'To avoid this behavior, consider changing the starting point, bounds, or value of'
                                       'NM_StartingSimplex')
                else:
                    y[k] = zdelt
                sim[k + 1] = y
        elif isinstance(self.NM_StartingSimplex, np.ndarray)or isinstance(self.NM_StartingSimplex, list):
            sim = self.NM_StartingSimplex
            # then we assume the user entered their own simplex. No additional error handling is included because
            # scipy do their own
        else:
            logger.error('Starting simplex can only be defined as a relative parameter, e.g. 0.1, or as an array of size'
                         '[N+1, N]. Quitting')
            sys.exit(1)

        self.StartingSimplex = sim

    def RunOptimisation(self):
        """
        Use the scipy.optimize.minimize module to perform the optimisation.
        Note that most of the 'action' is happening in BlackBoxFunction, which is repeated called by the optimiser
        """

        self.SetUpDirectoryStructure()
        if self.StartingSimplexSupplied:
            StartingSimplex = self.StartingSimplex
        else:
            StartingSimplex = None

        bnds = tuple(zip(self.LowerBounds, self.UpperBounds))

        self.NelderMeadRes = minimize(self.BlackBoxFunction, self.StartingValues, method='Nelder-Mead', bounds=bnds,
                       options={'disp': True, 'initial_simplex': StartingSimplex,
                                'maxiter': self.MaxItterations, 'maxfev': self.MaxItterations})
        self._write_final_log_entry()


class BayesianOptimiser(TopasOptBaseClass):
    """
    Class to perform optimisation using the `Bayesian Optimisation code <https://github.com/fmfn/BayesianOptimization>`_
    Specific options are described below, the rest are described in TopasOptBaseClass

    :param bayes_length_scales: Bayes-specific parameter to defined the length scales used in the gaussian process
        model. If supplied to non-Bayes optimisers it does nothing (and hopefully you will be warned accordingly).
        This can be supplied as one of three things: **None**: in this case, the default is used: length_scale=1.0
        **Number > 0 and <1**: in this case, the length scales for each parameter are derived as a percentage of range.
        For instance if the user enter 0.1, all length scales will be set to 10% of the range of each variable -
        this is the default behavior. **Array**: Finally, the user is free to simply specify what length scales to use for each parameter. Make sure
        you enter them in alphabetical order as this is the order used internally by the optimiser.
    :type bayes_length_scales: None, float, or array (optional)
    :param bayes_UCBkappa: Bayes-specific parameter . kappa value in UCB  function. A higher value=more exploration. see
        `this notebook <https://github.com/fmfn/BayesianOptimization/blob/master/examples/exploitation_vs_exploration.ipynb>`_
        for explanation
    :type bayes_UCBkappa: float, optional
    :param bayes_KappaDecayIterations: Bayes-specific parameter. Over the last N iterations, kappa will decay to be
        almost 0 (highly exploitive). For explantion of kappa decay see `here <https://github.com/fmfn/BayesianOptimization/pull/221>`_
    :type bayes_KappaDecayIterations: int, optional
    :param bayes_GP_alpha: Bayes-specific parameter. This parameter handles the smoothness of the gaussian process model.
        for a noisy objective function, increasing this value can minimise overfitting errors.
    :type bayes_GP_alpha: float, optional
    """

    def __init__(self, bayes_length_scales=None, bayes_UCBkappa=5,
                 bayes_KappaDecayIterations=10, bayes_GP_alpha=0.01, **kwds):
        """
        init function for Bayesian optimiser
        """
        self.bayes_length_scales = bayes_length_scales
        self.bayes_UCBkappa = bayes_UCBkappa
        self.bayes_KappaDecayIterations = bayes_KappaDecayIterations
        self.bayes_GP_alpha = bayes_GP_alpha
        super().__init__(**kwds)

        self.BayesOptLogLoc = Path(self.BaseDirectory) / self.SimulationName / 'logs/bayes_opt_logs.json'
        self._BayesianOptimiser__RestartMode = False  # don't change!
        self._create_p_bounds(self._optimisation_params)  # Bayesian optimiser wants bounds in a slight differnt format
        self._create_suggested_points_to_probe(self._optimisation_params)
        self._derive_bayes_length_scales(bayes_length_scales)

        # Bayesian optimisation settings:
        self._target_prediction_mean = []  # keep track of what the optimiser expects to get
        self._target_prediction_std = []  # keep track of what the optimiser expects to get
        # see here: https://github.com/fmfn/BayesianOptimization/issues/202
        self.Matern_Nu = 1.5  # see here https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html#sklearn.gaussian_process.kernels.Matern
        self.UCBkappa = bayes_UCBkappa  # higher kappa = more exploration. lower kappa = more exploitation
        self.n_restarts_optimizer = 20  # this controls the gaussian process fitting. 20 seems to be a good number.
        self.bayes_KappaDecayIterations = bayes_KappaDecayIterations
        self.UCBKappa_final = 0.1
        self.kappa_decay_delay = self.MaxItterations - self.bayes_KappaDecayIterations  # this many exploritive iterations will be carried out before kappa begins to decay

        if self.kappa_decay_delay >= self.MaxItterations:
            logger.warning(f'Kappa decay requested, but since kappa_decay_delay ({self.kappa_decay_delay}) is less'
                           f'than MaxItterations ({self.MaxItterations}), decay will never occur...')
            self.kappa_decay = 1
        else:
            self.kappa_decay = (self.UCBKappa_final / self.UCBkappa) ** (
                        1 / (self.MaxItterations - self.kappa_decay_delay))
            # ^^ this is the parameter to ensure we end up with UCBKappa_final on the last iteration


    def _derive_bayes_length_scales(self, bayes_length_scales):
        """
        Figure out what to put in to the gaussian process model kernel for length scales.
        We use the Matern Kernel which is detailed `here <https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html#sklearn.gaussian_process.kernels.Matern>`_

        :param bayes_length_scales: user input for length scales. Can be one of three things:
            1. **None**: in this case, the default is used: length_scale=1.0
            2. Number > 0 and <1: in this case, the length scales for each parameter are derived as a percentage of range.
                For instance if the user enter 0.1, all length scales will be set to 10% of the range of each variable.
            3. Array: Finally, the user is free to simply specify what length scales to use for each parameter. Make sure
                you enter them in alphabetical order as this is the order used internally by the optimiser.
        """
        ParameterNames = sorted(self.pbounds.keys())
        if bayes_length_scales is None:
            bayes_length_scales = 0.1
            self.bayes_length_scales = []
            for paramter_name in ParameterNames:
                length_scale_temp = (self.pbounds[paramter_name][1] - self.pbounds[paramter_name][0]) * bayes_length_scales
                self.bayes_length_scales.append(length_scale_temp)
        elif type(bayes_length_scales) is float:
            self.bayes_length_scales = []
            for paramter_name in ParameterNames:
                length_scale_temp = (self.pbounds[paramter_name][1] - self.pbounds[paramter_name][0]) * bayes_length_scales
                self.bayes_length_scales.append(length_scale_temp)
        elif type(bayes_length_scales) is list or type(bayes_length_scales) is np.ndarray:

            if not len(bayes_length_scales) == len(ParameterNames):
                logger.error(f'length of bayes_length_scales must be single values or match number of parameters; you have'
                             f'{len(ParameterNames)} Parameters, but {len(bayes_length_scales)} bayes_length_scales. Quitting')
                sys.exit(1)
            self.bayes_length_scales = bayes_length_scales

    def _create_p_bounds(self, optimisation_params):
        """
        Just reformat the optimisation variables into a format that the Bayesian optimiser wants
        """

        pbounds = {}
        for i, ParamName in enumerate(optimisation_params['ParameterNames']):
            pbounds[ParamName] = (optimisation_params['LowerBounds'][i], optimisation_params['UpperBounds'][i])

        self.pbounds = pbounds

    def _create_suggested_points_to_probe(self, optimisation_params):
        """
        If the user has entered any suggestions, sort them alphabetically and create some variables to keep track of them
        """
        if not 'Suggestions' in optimisation_params.keys():
            # no suggestions made and that's fine
            self.Nsuggestions = None
            return

        else:
            if optimisation_params['Suggestions'].ndim == 1:
                # add an empty dimension to keep the flow going
                optimisation_params['Suggestions'] = np.expand_dims(optimisation_params['Suggestions'], 0)
            if not optimisation_params['Suggestions'].shape[1] == len(optimisation_params['ParameterNames']):
                logger.error(f"wrong dimensionality for suggestions: {optimisation_params['Suggestions'].shape}."
                             f"\nSuggestions shape should be N*{len(optimisation_params['ParameterNames'])}")
                sys.exit(1)
            self.Nsuggestions = optimisation_params['Suggestions'].shape[0]
            self.Suggestions = optimisation_params['Suggestions']
            # put the suggestions into the correct order and store in an array
            self.Suggestions = []
            for n in range(self.Nsuggestions):
                suggestion_dict = {}
                for i, key in enumerate(optimisation_params['ParameterNames']):
                    suggestion_dict[key] = optimisation_params['Suggestions'][n, i]
                self.Suggestions.append(suggestion_dict)

    def _plot_convergence_plot_retrospective(self, optimizer):
        """
        This will read in a log file and produce a version of the convergence plot using the input gaussian process
        model to predict the objective function at each itteration.
        This allows a comparison between the prospective and retrospective performance of the GPM. the plot produced
        in real time (SphinxOptBaseClass._Plot_Convergence) will show the performance of the model in 'real time'.
         this will show the performance
        of the final model.

        :param optimizer: The bayesian optimiser object from RunOptimisation
        """

        ResultsDict = ReadInLogFile(self._LogFileLoc)

        # we need to format this into a an array based on ParameterNames
        ResultsArray = np.zeros([len(ResultsDict['Itteration']), len(self.ParameterNames)])

        SortedNames = sorted(self.ParameterNames)  # this took me actually hours to figure out.
        for i, parameter in enumerate(SortedNames):
            Values = ResultsDict[parameter]
            ResultsArray[:, i] = Values

        OF = ResultsDict['ObjectiveFunction']
        OF = [-1 * item for item in
              OF]  # it is stored in the log files with the opposite sign of what is used internally
        ItterationVector = ResultsDict['Itteration']
        mean, std = optimizer._gp.predict(ResultsArray, return_std=True)

        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(5, 5))
        axs.plot(ItterationVector, OF, 'C1-x')
        axs.set_xlabel('Itteration number', fontsize=FigureSpecs.LabelFontSize)
        axs.set_ylabel('Objective function', fontsize=FigureSpecs.LabelFontSize)
        axs.grid(True)
        axs.plot(ItterationVector, mean, 'C0')
        axs.fill_between(ItterationVector, mean + std, mean - std, alpha=0.3, color='C0')
        axs.legend(['Actual', 'Predicted', 'Std.'])
        axs.set_title('Retrospective Model Fit', fontsize=FigureSpecs.TitleFontSize)

        SaveLoc = Path(self.BaseDirectory) / self.SimulationName
        SaveLoc = SaveLoc / 'logs'
        plt.savefig(SaveLoc / 'RetrospectiveModelFit.png')
        plt.close(fig)

    def _plot_predicted_versus_actual_correlation(self):
        """
        Produce a scatter plot of the predicted versus actual objective function values, and print spearman and pearson
        correlation coefficient to it
        """
        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(5, 5))
        TargetPredictionMean = [-1 * item for item in
                                self._target_prediction_mean]  # need to account for min/max discrepancy
        axs.scatter(self.AllObjectiveFunctionValues, TargetPredictionMean)
        axs.set_xlabel('Actual', fontsize=FigureSpecs.LabelFontSize)
        axs.set_ylabel('Predicted', fontsize=FigureSpecs.LabelFontSize)
        NewMinLim = np.min([axs.get_ylim(), axs.get_xlim()])
        NewMaxLim = np.max([axs.get_ylim(), axs.get_xlim()])
        axs.set_ylim([NewMinLim, NewMaxLim])
        axs.set_xlim([NewMinLim, NewMaxLim])
        axs.plot([NewMinLim, NewMaxLim], [NewMinLim, NewMaxLim], 'r--')

        # Assess and print correlation metrics:
        if len(TargetPredictionMean) > 2:
            Pearson = stats.pearsonr(self.AllObjectiveFunctionValues, TargetPredictionMean)
            Spearman = stats.spearmanr(self.AllObjectiveFunctionValues, TargetPredictionMean)
            plt.text(NewMinLim + abs(NewMinLim * 0.2), NewMaxLim - (abs(NewMaxLim * 0.2)),
                     f'Spearman: {Spearman[0]: 1.1f}\nPearson: {Pearson[0]: 1.1f}',
                     fontsize=FigureSpecs.LabelFontSize)

        axs.grid(True)
        axs.set_title('Actual versus predicted correlation', fontsize=FigureSpecs.TitleFontSize)
        plt.tight_layout()
        SaveLoc = Path(self.BaseDirectory) / self.SimulationName
        SaveLoc = SaveLoc / 'logs'
        plt.savefig(SaveLoc / 'CorrelationPlot.png')
        plt.close(fig)

    def _plot_single_variable_objective(self, optimizer):
        """
        For each variables, produce a plot of the prediced objective function while only that parameter varies
        and the others are held at the current best values.
        This allows one to get an idea of the predicted objective function shape. The accuracy of this prediction
        depends on how well the model is currently fit...
        """

        PlotSavePath = str(Path(self.BaseDirectory) / self.SimulationName / 'logs' / 'SingleParameterPlots')
        if not os.path.isdir(PlotSavePath):
            os.mkdir(PlotSavePath)

        Nsamples = 100
        Params = list(optimizer.max['params'].keys())
        PointsToTest = np.ones([len(Params), Nsamples])  # this will be an array of the best points for each param
        # below we overwrite one row at a time with the parameter being varied
        ConstantValues = np.array(list(optimizer.max['params'].values()))  # current best guess for each param
        for i, const_val in enumerate(ConstantValues):
            PointsToTest[i, :] = PointsToTest[i,:] * const_val


        for i, param in enumerate(Params):

            PointsToTest_temp = PointsToTest.copy()
            PointsToVary = np.linspace(self.pbounds[param][0], self.pbounds[param][1], Nsamples)
            PointsToTest_temp[i, :] = PointsToVary
            mean, std = optimizer._gp.predict(PointsToTest_temp.T, return_std=True)
            mean = -1 * mean
            std = -1 * std
            fig = plt.figure()
            plt.plot(PointsToVary, mean)
            plt.fill_between(PointsToVary, mean + std, mean - std, alpha=0.5, color='C0')
            plt.xlabel(param)
            plt.ylabel('Predicted Objective function value')
            plt.title(f'Single parameter plot for {param}')
            plt.grid()
            SaveName = PlotSavePath + f'/{param}.png'
            plt.savefig(SaveName)
            plt.close(fig)

    def _update_logs_with_length_scales(self):
        """
        writes the original and optimised length scales to the end of a log file
        """
        start_length = self.optimizer._gp.kernel.length_scale
        final_length = self.optimizer._gp.kernel_.length_scale
        keys = sorted(self.pbounds)
        start_length = dict(zip(keys, start_length))
        final_length = dict(zip(keys, final_length))
        Entry = f'\nStarting length scales were {start_length}'
        Entry = Entry + f'\nOptimised length scales were {final_length}'

        with open(self._LogFileLoc , 'a') as f:
            f.write(Entry)

    def RunOptimisation(self):
        """
        This is the main optimisation loop.
        For explanation of the various parameters and commands, start with the
        """

        # set up directory structure
        if not self.__RestartMode:
            self.SetUpDirectoryStructure()
        else:
            self._setup_topas_emulator()

        # instantiate optimizer:

        self.optimizer = BayesianOptimization(f=None, pbounds=self.pbounds, random_state=1)
        self.optimizer.set_gp_params(normalize_y=True, kernel=Matern(length_scale=self.bayes_length_scales, nu=self.Matern_Nu),
                                n_restarts_optimizer=self.n_restarts_optimizer, alpha=self.bayes_GP_alpha)  # tuning of the gaussian parameters...
        utility = UtilityFunction(kind="ucb", kappa=self.UCBkappa, xi=0.0, kappa_decay_delay=self.kappa_decay_delay,
                                  kappa_decay=self.kappa_decay)
        logger.warning('setting numpy to ignore underflow errors.')
        np.seterr(under='ignore')

        if self.__RestartMode:
            # then load the previous log files:
            load_logs(self.optimizer, logs=[self.BayesOptLogLoc])
            bayes_opt_logger = newJSONLogger(path=str(self.BayesOptLogLoc))
            self.optimizer.subscribe(Events.OPTIMIZATION_STEP, bayes_opt_logger)
            self.optimizer._gp.fit(self.optimizer._space.params, self.optimizer._space.target)
            self.Itteration = len(self.optimizer.space.target)
            self.ItterationStart = len(self.optimizer.space.target)
            utility._iters_counter = self.ItterationStart
            if self.Itteration >= self.MaxItterations-1:
                logger.error(f'nothing to restart; max iterations is {self.MaxItterations} and have already been completed')
                sys.exit(1)
        else:
            bayes_opt_logger = JSONLogger(path=str(self.BayesOptLogLoc))
            self.optimizer.subscribe(Events.OPTIMIZATION_STEP, bayes_opt_logger)
            # first guess is nonsense but we need the vectors to be the same length
            self._target_prediction_mean.append(0)
            self._target_prediction_std.append(0)
            target = self.BlackBoxFunction(self.VariableDict)
            self.optimizer.register(self.VariableDict, target=target)

        for point in range(self.Itteration, self.MaxItterations):
            utility.update_params()
            if (self.Nsuggestions is not None) and (self.SuggestionsProbed < self.Nsuggestions):
                # evaluate any suggested solutions first
                next_point_to_probe = self.Suggestions[self.SuggestionsProbed]
                self.SuggestionsProbed += 1
            else:
                next_point_to_probe = self.optimizer.suggest(utility)

            NextPointValues = np.array(list(next_point_to_probe.values()))
            mean, std = self.optimizer._gp.predict(NextPointValues.reshape(1, -1), return_std=True)
            self._target_prediction_mean.append(mean[0])
            self._target_prediction_std.append(std[0])
            target = self.BlackBoxFunction(next_point_to_probe)
            try:
                self.optimizer.register(params=next_point_to_probe, target=target)
            except KeyError:
                try:
                    self.RepeatedPointsProbed = self.RepeatedPointsProbed + 1
                    logger.warning(
                        f'Bayesian algorithm is attempting to probe an existing point: {NextPointValues}. Continuing for now....')
                    if self.RepeatedPointsProbed > 10:
                        logger.error('The same point has been requested more than 10 times; quitting')
                        break
                except AttributeError:
                    self.RepeatedPointsProbed = 1

            self._plot_predicted_versus_actual_correlation()
            self._plot_convergence_plot_retrospective(self.optimizer)
            self._plot_single_variable_objective(self.optimizer)

        # update the logs with the best value:
        self._write_final_log_entry()
        # update logs with length scales:
        self._update_logs_with_length_scales()

    def RestartOptimisation(self):
        """
        Sometimes for whatever reason an optimisation is stopped prematurely.
        This function allows you to restart the optimisation by loading the previous log files.
        You just have to change Optimiser.RunOptimisation() to Optimiser.RestartOptimisation()
        in your optimisation script; the code will do the rest automatically.
        """
        self.__RestartMode = True


        # delete the end of run info from the log file
        with open(self._LogFileLoc, "r") as f:
            lines = f.readlines()
        with open(self._LogFileLoc, "w") as f:
            for line in lines:
                if line[0:10] == 'Itteration':
                    f.write(line)

        self.RunOptimisation()




"""
Convert topas parameter files into a python function that recreates the input files.
"""
import shutil
import sys, os
import logging
from pathlib import Path
import re

logger = logging.getLogger(__name__)
logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')


class generate_topas_script_generator:
    """
    This code will take a list topas scripts, and create a python function that returns a list for each script.
    Each list element contains one line of the topas script, which can be used to automatically regenerate that script.
    This class will also attempt to handle input/output lines in the topas language such as include lines and input
    and output of phase space lines.

    :param OutputDirectory: location to write the returned python function to
    :type OutputDirectory: string or pathlib.Path
    :param TopasScriptLocation: location of all topas scripts which need to be recreated. Note that you do not need
        to enter 'include' files, as these are found automatically. You only need to enter files which will be
        changing on each iteration. The order of the files must match the order in which they will be run so the
        code can correctly detect dynamic input/output files.
    :type TopasScriptLocation: string or pathlib.Path
    :param IncludeFileStorageDirectory: Any included files will be copied here. This is to try and ensure the topas
        scripts are portable. If no path is entered then they will be written to OutputDirectory / IncludeFiles
    :type IncludeFileStorageDirectory: None or string or pathlib.Path (optional)
    :param ErrorChecking: if True (recommended), will attempt to perform some checks on the input data.
    :type ErrorChecking: bool, optional
    """

    def __init__(self,OutputDirectory, TopasScriptLocation, IncludeFileStorageDirectory=None,
                 ErrorChecking=True):
        """
        init method TopasScriptGenerator. input options are in class docstring
        """

        self.OutputDirectory = Path(OutputDirectory)
        self.TopasScriptLocation = TopasScriptLocation


        if IncludeFileStorageDirectory is None:
            self.IncludeFileStorageDirectory = Path(OutputDirectory) / 'IncludeFiles'

        else:
            self.IncludeFileStorageDirectory = IncludeFileStorageDirectory

        self.ErrorChecking = ErrorChecking
        if self.ErrorChecking:
            self._check_inputs()
        self._generate_topas_script_generator()

    def _check_inputs(self):
        """
        Perform checks of input data.
        these tests should be updated as uncaught errors arise!
        """

        if not type(self.TopasScriptLocation) is list:
            self.TopasScriptLocation = [self.TopasScriptLocation]

        for script in self.TopasScriptLocation:
            if not os.path.isfile(script):
                logger.error(f'Could not find TopasScriptLocation at {script}. Quitting')
                sys.exit(1)

        if not os.path.isdir(self.OutputDirectory):
            logger.error(f'Specified output directory does not exist: {self.OutputDirectory}')


    def _handle_include_files(self, OriginalFileLocation, line):
        """
        This function is called when an includeFile statement is found in the input topas script.
        The include file will be copied to  IncludeFileStorageDirectory, and the line updated with an
        absolute path to that location

        Returns: line, the updated includeFile line
        """
        # first, check that this include file exists... the user may have used a relative or an absolute path...
        # see if relative paths points to a OriginalFileLocation:
        if not os.path.isdir(self.IncludeFileStorageDirectory):
            os.mkdir(self.IncludeFileStorageDirectory)
        FileLocation, dum = os.path.split(OriginalFileLocation)
        IncludeFileNames = re.findall('(?<==).*$', line)[0] # extract the file name with regular expression
        IncludeFileNames = IncludeFileNames.split('#')[0]  # ignore anything after a comment symbol
        IncludeFileNames = IncludeFileNames.split(' ')
        line = 'includeFile = '
        for IncludeFileName in IncludeFileNames:  # have to handle the potential for multiple include files on one line
            if IncludeFileName == '' or IncludeFileName == '\n':
                continue
            IncludeFileName = IncludeFileName.replace(' ','')  # replace any spaces with strings
            if os.path.isfile(FileLocation + '/' + IncludeFileName):
                IncludeFileLocation = Path(FileLocation) / IncludeFileName
            elif os.path.isfile(IncludeFileName):  # check for use with absolute paths
                IncludeFileLocation = Path(FileLocation) / IncludeFileName
                dum, IncludeFileName = os.path.split(IncludeFileName)
            else:
                logger.error(f'could not find include file defined on this line:\n{line}'
                             f'\nSearched locations:'
                             f'\n{FileLocation + "/" + IncludeFileName}'
                             f'\n{IncludeFileName}'
                             f'\nQuitting')
                sys.exit(1)

            # assuming we found the file, we have to 1) copy it
            shutil.copy2(IncludeFileLocation, str(self.IncludeFileStorageDirectory) + '/' + IncludeFileName)
            # 2) check if it itself contains any include files! This happens recursively (this function calls itself)
            f = open(IncludeFileLocation)
            for i, include_file_line in enumerate(f):
                if 'includeFile'.lower() in include_file_line.lower():
                    if not include_file_line.lstrip()[0] == '#':  # ignore comments
                        # extract everything on the RHS of the equals sign:
                        RHS = include_file_line.lstrip().split('=')[1]
                        RHS = RHS.split('#')[0]  #remove anything after a comment symbol
                        AllIncludeFiles = RHS.split(' ')
                        new_include_file_line = 'includeFile = '
                        for j, file in enumerate(AllIncludeFiles):
                            if file == '' or file =='\n':
                                continue
                            temp_include_file_line = 'includeFile = ' + file
                            include_OriginalFileLocation_line = self._handle_include_files(IncludeFileLocation, temp_include_file_line)
                            new_include_file_line = new_include_file_line + os.path.split(self.IncludeFileStorageDirectory)[1] + '/' + file + ' '
                        if AllIncludeFiles[-1] == '\n':
                            new_include_file_line = new_include_file_line + '\n'
                        # now we need to update this line in the copied script:
                        copied_file_location = str(self.IncludeFileStorageDirectory) + '/' + IncludeFileName
                        f2 = open(copied_file_location, 'r')
                        old_lines = f2.readlines()
                        new_lines = old_lines.copy()
                        new_lines[i] = new_include_file_line
                        f2 = open(copied_file_location, 'w')
                        f2.writelines(new_lines)

            # 3) update the line to point to the new storage location.
            line = line + ' ' + os.path.split(self.IncludeFileStorageDirectory)[1] + '/' + IncludeFileName

        return line

    def _handle_output_file_line(self, line):
        """
        If a file is being output, change the location it is being output to
        BaseDirectory / Results.
        Change the name of the output to reflect the iteration number.

        Args:
            line: the OutputFile line from the original

        Returns:
            line: the line adapted to output a file name to BaseDirectory / Results, and with the relevat iteration
                  appended

        """
        line1, line2 = line.split("=", 1)  # delete everything after the first =
        dum, OriginalFileName = line2.split("/", 1)
        OriginalFileName = OriginalFileName.replace('"', '')
        OriginalFileName = OriginalFileName.replace("'", '')
        line1 = line1.replace('"', '')
        line1 = line1.replace("'", '')
        new_line = line1 + ' =  "../Results/' + OriginalFileName + "_itt_\' + " + 'str(iteration)' + ' + \'"\')'

        return new_line, OriginalFileName

    def _handle_phase_space_source(self,OriginalFileLocation, line, OutputPhaseSpaceFilesNames):
        """
        Handle lines where a phase space source is used

        For phase space sources, there are two scenarios:
        1. A static phase space source is used; in this case we just want to update

        Args:
            line:
            OutputPhaseSpaceFilesNames:

        Returns:

        """
        # check whether this is a dynamic or static phase space source
        UseStaticPhaseSpace = True
        for OutputPhaseSpace in OutputPhaseSpaceFilesNames:
            if OutputPhaseSpace in line:
                UseStaticPhaseSpace = False

        line1, line2 = line.split("=", 1)  # delete everything after the first =
        OriginalFileName = line2.split('/')[-1]
        OriginalFileName = OriginalFileName.replace('"', '')
        OriginalFileName = OriginalFileName.replace("'", '')
        OriginalFileName = OriginalFileName.replace(" ", '')
        line1 = line1.replace('"', '')
        line1 = line1.replace("'", '')
        if not UseStaticPhaseSpace:
            # dynamically allocated phase space name
            # new_line = line1 + ' = "\' + str(Path(BaseDirectory) / "Results" / "' + OriginalFileName + '") + "_itt" + ' + 'str(iteration)' + ' + \'"\')'
            new_line = line1 + ' =  "../Results/' + OriginalFileName + "_itt_\' + " + 'str(iteration)' + ' + \'"\')'
        else:
            # in this case we want to convert a relative path to an absolute path.
            # do we also want to copy the file to the IncludeDirectory? probably not, it could be huge...
            new_line = line1 + ' = ' + str(OriginalFileLocation.parent / OriginalFileName) + "')"

        return new_line

    def _generate_topas_script_generator(self):
        """
        generate a python script that will itself generate the script at TopasScriptLocation
        """
        outputFile = self.OutputDirectory / 'GenerateTopasScripts.py'
        # first read in the text:

        TopasScriptGeneratorReadMe = '    """\n    This file simply returns a list object, where each list entry corresponds to' \
                                     '\n    a line in the topas script.\n    When it is called from an Optimiser object,' \
                                     'it will receive a dictionary that contains the current values of \n    ' \
                                     'the variables you set up in optimisation_params when you initialised the optimiser.\n    """\n'

        TopasScriptGenerator = []
        TopasScriptGenerator.append('from pathlib import Path\n')
        TopasScriptGenerator.append('\n')
        TopasScriptGenerator.append('def GenerateTopasScripts(BaseDirectory, iteration, **variable_dict):\n')
        TopasScriptGenerator.append(TopasScriptGeneratorReadMe)

        ScriptNames = []
        ReturnStatementString = '['
        OutputPhaseSpaceFilesNames = []  # need to keep track of these to see if they match the input of another file
        ScriptNames = []
        for file in self.TopasScriptLocation:
            dum, ScriptName = os.path.split(file)
            ScriptName, dum = os.path.splitext(ScriptName)
            ScriptName = ''.join(e for e in ScriptName if e.isalnum())  # heal any weird chracters
            ReturnStatementString = ReturnStatementString + ScriptName + ', '
            ScriptNames.append(ScriptName)
            f = open(file)
            TopasScriptGenerator.append('    \n')
            TopasScriptGenerator.append('    ' + ScriptName + ' = []\n')

            for line in f:
                line = line.replace('\n', '')  # remove new line characters
                CommentLine = False
                if (not line == '') and line.lstrip()[0] == '#':
                    CommentLine = True

                if not CommentLine:  # avoid doing anything silly to commented lines
                    line = line.split('#', 1)
                    line = line[0]  # i'm going to remove inline comments, easier and safer.
                    if 'includeFile'.lower() in line.lower():
                        line = self._handle_include_files(file, line)
                    if 'OutputFile '.lower() in line.lower():
                        line, OriginalFileName = self._handle_output_file_line(line)
                        TopasScriptGenerator.append("    " + ScriptName + ".append('" + line + "\n")
                        OutputPhaseSpaceFilesNames.append(OriginalFileName)
                        continue  # need a slightly different line in this case from the default
                    if 'PhaseSpaceFileName'.lower() in line.lower():  # could put a more sophisticated test here...
                        line = self._handle_phase_space_source(file, line, OutputPhaseSpaceFilesNames)
                        TopasScriptGenerator.append("    " + ScriptName + ".append('" + line + "\n")
                        continue  # need a slightly different line in this case from the default

                TopasScriptGenerator.append("    " + ScriptName + ".append('" + line + "')\n")

        ReturnStatementString = ReturnStatementString[:-2]  # remove the last comma
        ReturnStatementString = ReturnStatementString + ']'
        ReturnStatementString = ReturnStatementString + ', ' + str(ScriptNames)
        TopasScriptGenerator.append('\n    return ' + ReturnStatementString)

        TopasScriptGenerator.append('\n\nif __name__ == "__main__":\n')
        TopasScriptGenerator.append('    Scripts, ScriptNames = GenerateTopasScripts(".", 1)\n')
        TopasScriptGenerator.append('    for i, script in enumerate(Scripts):\n')
        TopasScriptGenerator.append('        filename = ScriptNames[i] + ".tps"\n')
        TopasScriptGenerator.append('        f = open(filename, "w")\n')
        TopasScriptGenerator.append('        for line in script:\n')
        TopasScriptGenerator.append('            f.write(line)\n')
        TopasScriptGenerator.append('            f.write("\\n")\n')
        f2 = open(outputFile, "w+")
        for line in TopasScriptGenerator:
            f2.writelines(line)


"""
TopasOpt is a library for performing inverse optimisation with Topas Monte Carlo.
Please see the docs page for more info:
https://acrf-image-x-institute.github.io/TopasOpt/
"""
__author__ = 'Brendan Whelan'
__version__ = '0.3.0'





"""
Supporting classes and functions that don't belong anywhere else in particular
"""
from bayes_opt.logger import JSONLogger
import sys
import os
sys.path.append('.')
# from WaterTankAnalyser import WaterTankData
# import numpy as np
# from matplotlib import pyplot as plt
import os, sys
import logging
import numpy as np
import topas2numpy as tp
import matplotlib.pyplot as plt
from scipy.interpolate import RegularGridInterpolator
from scipy.stats import linregress
from pathlib import Path
plt.interactive(False)

ch = logging.StreamHandler()
formatter = logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.INFO)  # This toggles all the logging in your app
logger.propagate = False

class bcolors:
    """
    This is just here to enable me to print pretty colors to the linux terminal
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class FigureSpecs:
    """
    Thought this might be the easiest way to ensure universal parameters accross all figures
    """
    LabelFontSize = 14
    TitleFontSize = 16
    Font = 'serif'
    AxisFontSize = 14


class newJSONLogger(JSONLogger):
    """
    To avoid the annoying behaviour where the bayesian logs get deleted on restart.
    Thanks to: https://github.com/fmfn/BayesianOptimization/issues/159
    """
    def __init__(self, path):
        """
        I have to put a docstring here because I can't figure out how to make interrogate ignore it :-/
        """
        self._path = None
        super(JSONLogger, self).__init__()
        self._path = path if path[-5:] == ".json" else path + ".json"


class WaterTankData:
    """
    Read in and analyse a series of topas scoring files in a rectangular phantom (the water tank).
    There are two ways you might want to use this:

    1. Analyse a single dose file
    2. Analyse multiple dose files all on the same scoring grid

    The main attribute is the 'DoseCube' array. This is the agregate of all the input files. If you
    only put one file in, then this is simply the agregate of that single file.
    A number of plotting routines are already provided which operate on this array, but if you need to you can use
    ExtractDataFromDoseCube to generate any additional data from it.

    Basic use::

        home = os.path.expanduser("~")
        AnalysisPath = home + '/Dropbox (Sydney Uni)/Projects/PhaserSims/topas/Paper_7mm_all/Results/'
        FileToAnalyse = GetAllBinFiles(AnalysisPath)
        Dose = WaterTankData(AnalysisPath, FileToAnalyse, MirrorData=True)
        Dose.Plot_DosePlanes()
        # extract some additional data from agregate dose cube, XY plane for example:
        [Xpts, Ypts] = np.meshgrid(Dose.x, Dose.y)
        Zpts = Dose.PhantomSizeZ * np.ones(Xpts.shape)
        XY_data = Dose.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)

    """

    def __init__(self, AnalysisPath, FileToAnalyse, AbsDepthDose=False):
        """
        :param AnalysisPath: Path where files are located
        :type AnalysisPath: string
        :param FileToAnalyse: Can be either a single file or a list of tiles
        :type FileToAnalyse: list or string
        :param AbsDepthDose: If True, plots are in dose, if False they are normalised to Dmax
        :type AbsDepthDose: Boolean, optional
        """

        self.AnalysisPath = AnalysisPath
        self.FileToAnalyse = FileToAnalyse
        # the below lists all get appended to as data is read in
        self.Xangles = []
        self.Yangles = []
        self.MaxDoseAtIsoplane = []
        self.MaxDoseAtIsoplaneGauss = []
        self.BeamletWidthsManual = []
        self.BeamletWidthsGauss = []
        self.CrossChannelLeakageRatio = []
        self.DmaxToD100 = []
        self.RelOutOfFieldSurfDose = []
        self.DosePerCoulomb = []
        self.LegendNames = []
        self.OF = []

        self.ReverseZDirection = True  # True makes the PDDs go in the normal direction
        self.AbsDepthDose = AbsDepthDose  # if true, dose is plotted instead of percentage

        # these settings keep track of whether certain warnigs have already been thrown
        self.DistanceUnitMessage = False  # just to keeep track of messages (only want to throw once)
        self.MirrorWarning = False  # just to keeep track of messages (only want to throw once)
        self.UnderflowWarning = False
        self.DoseUnitsWarning = False  # just to keeep track of messages (only want to throw once)

        # set figure specs:

        plt.rc('font', family=FigureSpecs.Font)
        plt.rc('xtick', labelsize=FigureSpecs.AxisFontSize)
        plt.rc('ytick', labelsize=FigureSpecs.AxisFontSize)
        self._ReadDoseFilesIntoDoseCube()
        self.__GenerateDepthDoseData()
        self.__GenerateProfileData()

    def _ReadDoseFilesIntoDoseCube(self):
        """
        read all dose files into an integral 'dose cube' that can be queried later
        """
        self.MultiFileMode = True  # overwrite below if not
        if isinstance(self.FileToAnalyse, str):
            # convert to list so we can use the same read in...
            self.FileToAnalyse = list([self.FileToAnalyse])
            self.MultiFileMode = False

        for file in self.FileToAnalyse:

            self.CurrentFile = file
            # file read in and processing
            FileName, Filetype = os.path.splitext(self.CurrentFile)
            self.LegendNames.append(FileName)

            self.__ReadInDoseFiles()
            self.__ConstructCoordinateSystem()

            # generate the data for plots:
            self.__GenerateDoseCubeData()

    def __ReadInDoseFiles(self):
        """
        read in the dose file using topas2numpy
        """
        try:
            FileLocation = str(Path(self.AnalysisPath) / self.CurrentFile)
            if not os.path.isfile(FileLocation):
                FileLocation = FileLocation + '.bin'
                if not os.path.isfile(FileLocation):
                    logging.error(f'Could not locate file: \n{FileLocation}')
            self.dose = tp.BinnedResult(FileLocation)
            # convert dose units:
            if self.dose.unit == 'Gy':
                DoseConverter = 1e6
                self.dose.unit = '\u03BCGy'
                if not self.DoseUnitsWarning:
                    print(f'converting Gy to {self.dose.unit}')
                    self.DoseUnitsWarning = True
            elif self.dose.unit == 'mGy':
                DoseConverter = 1000
                self.dose.unit = '\u03BCGy'
                if not self.DoseUnitsWarning:
                    print(f'converting mGy to {self.dose.unit}')
            else:
                DoseConverter = 1
                logging.warning('unable to detect dose unit; reading in without conversion')
            self.dose.data['Sum'] = self.dose.data['Sum'] * DoseConverter

        except FileNotFoundError as e:
            logging.warning(f'{bcolors.FAIL}Could not find one of the input files or file header: {self.CurrentFile}{bcolors.ENDC}\n')
            raise e

        if np.max(self.dose.data['Sum'])<1e-16:
            logging.error(f'there is no data in the dose file {FileLocation}')

    def __ConstructCoordinateSystem(self):
        """
        Build a coordinate system based off the dose cube information.
        For multifile mode, the coordinate system is only created the first time this function is called,
        otherwise it checks that the phantom dimensions are the same for subsequent files, and will cause
        an error if not
        """
        # what unit is the file in?
        if self.dose.dimensions[0].unit == 'cm':
            UnitConverter = 10
            if not self.DistanceUnitMessage:
                print('converting cm to mm')
                self.DistanceUnitMessage = True
        elif self.dose.dimensions[0].unit == 'mm':
            UnitConverter = 1
        else:
            UnitConverter = 1
            logging.warning('unit could not be detected, reading in without converion')

        try:
            # if coordinate system does exist, make sure the current file matches it....
            assert self.PhantomSizeX == ((max(self.dose.dimensions[0].get_bin_centers()) + (
                        self.dose.dimensions[0].bin_width / 2)) / 2) * UnitConverter
            assert self.PhantomSizeY == ((max(self.dose.dimensions[1].get_bin_centers()) + (
                        self.dose.dimensions[1].bin_width / 2)) / 2) * UnitConverter
            assert self.PhantomSizeZ == ((max(self.dose.dimensions[2].get_bin_centers()) + (
                        self.dose.dimensions[2].bin_width / 2)) / 2) * UnitConverter
        except AttributeError:
            # If coordinate system doesn't exist, construct it
            self.PhantomSizeX = ((max(self.dose.dimensions[0].get_bin_centers()) + (
                        self.dose.dimensions[0].bin_width / 2)) / 2) * UnitConverter
            self.PhantomSizeY = ((max(self.dose.dimensions[1].get_bin_centers()) + (
                        self.dose.dimensions[1].bin_width / 2)) / 2) * UnitConverter
            self.PhantomSizeZ = ((max(self.dose.dimensions[2].get_bin_centers()) + (
                        self.dose.dimensions[2].bin_width / 2)) / 2) * UnitConverter
            self.CoordOffsetX = self.PhantomSizeX  # assumes that phantom is centered around X/Y
            self.CoordOffsetY = self.PhantomSizeY
            self.CoordOffsetZ = 0
            self.x = (self.dose.dimensions[0].get_bin_centers() * UnitConverter) - self.CoordOffsetX
            self.y = (self.dose.dimensions[1].get_bin_centers() * UnitConverter) - self.CoordOffsetY
            self.z = (self.dose.dimensions[2].get_bin_centers() * UnitConverter) - self.CoordOffsetZ
            self.VoxelSizeX = ((self.PhantomSizeX * 2) / self.dose.dimensions[0].n_bins)
            self.VoxelSizeY = ((self.PhantomSizeY * 2) / self.dose.dimensions[1].n_bins)
            self.VoxelSizeZ = ((self.PhantomSizeZ * 2) / self.dose.dimensions[2].n_bins)

        self.DataSize = np.array([self.x.shape[0], self.y.shape[0], self.z.shape[0]])
        dim_check_ind = self.DataSize < 2  # check for singleton dimensions
        if dim_check_ind.any():
            # singleton dimensions will be mirrored to recover 3D data
            axes_names = np.array(['X', 'Y', 'Z'])
            logging.warning(f'singletony dimenions detected; expanding data in {axes_names[dim_check_ind][0]} to recover 3D')
            if dim_check_ind[0]:
                self.x = np.array([-self.x, self.x])
            elif dim_check_ind[1]:
                y_original = self.y[0]
                self.y = np.array([y_original - self.PhantomSizeY ,y_original+ self.PhantomSizeY ])
            elif dim_check_ind[2]:
                self.z = np.array([-self.z, self.z])

    def __GenerateDoseCubeData(self):
        """
        Read in all the data into a dose cube so we can query it later

        at the moment, this is done without interpolation, so there is an explicit assumption
        that each read in file has the same coordinate system (which is true for every use case I can currently envisage)
        """
        try:
            test = self.DoseCube
        except AttributeError:
            # the first time this function is called the data won't exist yet so it must be created
            self.DoseCube = np.zeros([self.x.size, self.y.size, self.z.size])
        Dose3D = self.dose.data['Sum']
        if (self.DataSize < 2).any():
            # we also need to expand the dimensinos
            if self.DataSize[0] < 2:
                Dose3D = np.stack([Dose3D.squeeze(),Dose3D.squeeze()], 0)
            elif self.DataSize[1] < 2:
                Dose3D = np.stack([Dose3D.squeeze(),Dose3D.squeeze()], 1)
            elif self.DataSize[2] < 2:
                Dose3D = np.stack([Dose3D.squeeze(),Dose3D.squeeze()], 2)
        self.DoseCube = np.add(self.DoseCube, Dose3D)

    def __GenerateDepthDoseData(self):
        """
        Interpolate depth dose data from current dose cube
        """
        # get logical array for central axis...
        Zpts_dd = self.z
        Xpts_dd = np.zeros(Zpts_dd.shape)
        Ypts_dd = np.zeros(Zpts_dd.shape)
        self.DepthDose  = self.ExtractDataFromDoseCube(Xpts_dd, Ypts_dd, Zpts_dd)

        if not self.AbsDepthDose:
            # then normalise plots to dmax:
            dmax = np.max(self.DepthDose)
            try:
                self.DepthDose = self.DepthDose * 100 / dmax
            except FloatingPointError:
                logging.warning(f'you seem to have divded by zero for files {self.FileToAnalyse[-1]}, which indicates that there is no actual dose '
                                'in the dose file')

    def __GenerateProfileData(self):
        """
        generate the data for plotting profiles.
        Note that this operates on each beamlet, NOT on the integral Dose Cube data
        """
        Xpts_prof = self.x  # profile over entire X range
        Ypts_prof = np.zeros(Xpts_prof.shape)
        Zpts_prof = self.PhantomSizeZ * np.ones(Xpts_prof.shape)  # at the middle of the water tank
        self.ProfileDose_X = self.ExtractDataFromDoseCube(Xpts_prof, Ypts_prof, Zpts_prof)
        # Y profile
        Ypts_prof = self.y  # profile over entire X range
        Xpts_prof = np.zeros(Ypts_prof.shape)
        Zpts_prof = self.PhantomSizeZ * np.ones(Xpts_prof.shape)
        self.ProfileDose_Y = self.ExtractDataFromDoseCube(Xpts_prof, Ypts_prof, Zpts_prof)

        if not self.AbsDepthDose:
            try:
                self.ProfileDose_X = self.ProfileDose_X * 100/max(self.ProfileDose_X)
                self.ProfileDose_Y = self.ProfileDose_Y * 100 / max(self.ProfileDose_Y)
            except FloatingPointError:
                logging.warning(f'you seem to have divded by zero for files {self.FileToAnalyse[-1]}, which indicates that there is no actual dose '
                                'in the dose file')

    # Public methods:

    def ExtractDataFromDoseCube(self, Xpts, Ypts, Zpts):
        """
        Extract data from the dose cube at positions Xpts, Ypts, Zpts
        Each of these is an array or list, and they must be the same shape (they can be of any dimensionality as
        they are flattened inside the function). The data is returned in the same shape as the input points.
        """
        assert Xpts.shape == Ypts.shape == Zpts.shape
        InputShape = Xpts.shape
        # convert to array and flatten
        Xpts = np.array(Xpts).flatten()
        Ypts = np.array(Ypts).flatten()
        Zpts = np.array(Zpts).flatten()

        try:
            test = self.DoseCubeInterpolator
        except AttributeError:
            self.DoseCubeInterpolator = RegularGridInterpolator((self.x, self.y, self.z), self.DoseCube,
                                                                bounds_error=False, fill_value=0)
            # interpolation function
        pts = np.stack([Xpts, Ypts, Zpts])
        pts = pts.T  # take the tranpose to get the dimnesions right
        InterpolatedData = self.DoseCubeInterpolator(pts)
        InterpolatedData = np.reshape(InterpolatedData,InputShape)

        return InterpolatedData

    def Plot_DosePlanes(self, AddColorBar=False):
        """
        Use the DoseCube data to create a plot through each of the cardinal planes.
        """

        fig, (axs1, axs2, axs3) = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
        # for generating extent in plots:
        dx = (self.x[1] - self.x[0]) / 2.
        dy = (self.y[1] - self.y[0]) / 2.
        dz = (self.z[1] - self.z[0]) / 2.

        #XY data
        extent = [self.x[0] - dx, self.x[-1] + dx, self.y[0] - dy, self.y[-1] + dy]
        [Xpts, Ypts] = np.meshgrid(self.x, self.y)
        Zpts = self.PhantomSizeZ * np.ones(Xpts.shape)
        XY_data = self.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)

        XYim = axs1.imshow(XY_data, extent=extent)
        axs1.set_xlabel('X [mm]', fontsize=FigureSpecs.LabelFontSize)
        axs1.set_ylabel('Y [mm]', fontsize=FigureSpecs.LabelFontSize)

        # ZX
        axs1.set_title('a) XY plane', fontsize=FigureSpecs.TitleFontSize)
        # fig.colorbar(im)

        #ZX data
        extent = [self.x[0] - dx, self.x[-1] + dx, self.z[0] - dz, self.z[-1] + dz]
        [Xpts, Zpts] = np.meshgrid(self.x, self.z)
        Ypts = np.zeros(Xpts.shape)
        ZX_data = self.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)
        # ZX_data = np.flipud(ZX_data)
        ZXim = axs2.imshow(ZX_data, extent=extent)
        axs2.set_xlabel('X [mm]', fontsize=FigureSpecs.LabelFontSize)
        axs2.set_ylabel('Z [mm]', fontsize=FigureSpecs.LabelFontSize)
        #
        axs2.set_title('b) ZX plane', fontsize=FigureSpecs.TitleFontSize)

        #ZY data
        extent = [self.y[0] - dy, self.y[-1] + dy, self.z[0] - dz, self.z[-1] + dz]
        [Ypts, Zpts] = np.meshgrid(self.y, self.z)
        Xpts = np.zeros(Zpts.shape)
        ZY_data = self.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)
        # ZY_data = np.flipud(ZY_data.T)
        ZYim = axs3.imshow(ZY_data, extent=extent)
        axs3.set_xlabel('Y [mm]', fontsize=FigureSpecs.LabelFontSize)
        axs3.set_ylabel('Z [mm]', fontsize=FigureSpecs.LabelFontSize)

        axs3.set_title('c) ZY plane', fontsize=FigureSpecs.TitleFontSize)

        if AddColorBar:
            cbar = fig.colorbar(XYim, ax=axs1)
            cbar.set_label('\u03BCGy')
            cbar = fig.colorbar(ZXim, ax=axs2)
            cbar.set_label('\u03BCGy')
            cbar = fig.colorbar(ZYim, ax=axs3)
            cbar.set_label('\u03BCGy')

        plt.tight_layout()
        plt.show()

    def Plot_DepthDose(self):
        """
        Plot integral depth dose curves of all beamlets
        """
        try:
            test = self.integralPDD_axs
        except AttributeError:
            fig, self.integralPDD_axs = plt.subplots()
        Zpts = self.z
        Xpts = np.zeros(Zpts.shape)
        Ypts = np.zeros(Zpts.shape)
        dose_plot = self.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)
        if self.ReverseZDirection:
            dose_plot = np.flip(dose_plot)
        if not self.AbsDepthDose:
            # then normalise plots to dmax:
            self.DepthDmax = np.max(dose_plot)
            dose_plot = dose_plot * 100 / self.DepthDmax
            self.integralPDD_axs.set_ylabel('Dose [%]', fontsize=FigureSpecs.LabelFontSize)
            self.integralPDD_axs.set_title('Percentage Depth Dose', fontsize=FigureSpecs.TitleFontSize)
        else:
            self.integralPDD_axs.set_ylabel(f'Dose [{self.dose.unit}]', fontsize=FigureSpecs.LabelFontSize)
            self.integralPDD_axs.set_title('Absolute Depth Dose', fontsize=FigureSpecs.TitleFontSize)
        self.integralPDD_axs.plot(self.z, dose_plot,'-o')
        self.integralPDD_axs.grid(True)
        self.integralPDD_axs.set_xlabel('Depth [mm]', fontsize=FigureSpecs.LabelFontSize)
        plt.show()

    def Plot_Profiles(self, dir='X', Zpoints=None):
        """
        Plot profiles through integrated data. Can choose dir='X' or dir='Y'
        You can also optionally pass multiple Z points; otherwise Z is set to the middle of the phantom
        """
        fig, IntProfile_axs = plt.subplots(figsize=(5, 5))
        if Zpoints is None:
            Zpoints = list([self.PhantomSizeZ])
        if not isinstance(Zpoints, list):
            Zpoints = list([Zpoints])
        if not (dir == 'X' or dir == 'Y'):
            logging.warning(f'invalid direction "{dir}" supplied PlotProfiles; valid options are X or Y. Picking X')
            dir = 'X'
        if dir == 'X':
            Xpts = self.x
            Ypts = np.zeros(Xpts.shape)
            AbcissaPlot = Xpts
            AbcissaLabel = 'X [mm]'
        elif dir == 'Y':
            Ypts = self.y
            Xpts = np.zeros(Ypts.shape)
            AbcissaPlot = Ypts
            AbcissaLabel = 'Y [mm]'

        leg_strings = []
        for zpoint in Zpoints:
            Zpts = zpoint * np.ones(Xpts.shape)
            ProfileData = self.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)
            leg_strings.append(f'Z = {zpoint} mm')
            IntProfile_axs.plot(AbcissaPlot, ProfileData, '-x', markersize=4)

        IntProfile_axs.set_xlabel(AbcissaLabel, size=FigureSpecs.LabelFontSize)
        IntProfile_axs.set_title('Beamlet Profiles', size=FigureSpecs.TitleFontSize)
        IntProfile_axs.set_ylabel(f'Dose [{self.dose.unit}]', size=FigureSpecs.LabelFontSize)
        IntProfile_axs.legend(leg_strings)
        IntProfile_axs.grid('True')
        plt.tight_layout()
        plt.show()

    def Plot_XYsurface(self, Zpoints=None):
        """
        Plot XY plots for multiple Z
        if Z = None, a single plot through the middle of the water phantom is produced
        """
        def GetSubplotsObject(Npoints):
            """
            Just returns an appropriate number of subplots and figure shape for the number of plots the user requested
            """
            if Npoints == 1:
                fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(5, 5))
                axs = np.array([axs])  # to make it match other cases
            elif Npoints == 2:
                fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))
            elif Npoints == 3:
                fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))
            elif Npoints == 4:
                fig, axs = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))
            elif Npoints == 5 or Npoints == 6:
                fig, axs = plt.subplots(nrows=3, ncols=3, figsize=(15, 10))
            else:
                logging.error("Too many Zpoints requested! you've ruined everything by being greedy!")
                sys.exit(1)
            return fig, axs

        if Zpoints is None:
            Zpoints = list([self.PhantomSizeZ])
        if not isinstance(Zpoints, list):
            Zpoints = list([Zpoints])
        figure, axs = GetSubplotsObject(len(Zpoints))
        leg_strings = []

        dx = (self.x[1] - self.x[0]) / 2.
        dy = (self.y[1] - self.y[0]) / 2.
        extent = [self.x[0] - dx, self.x[-1] + dx, self.y[0] - dy, self.y[-1] + dy]
        [Xpts, Ypts] = np.meshgrid(self.x, self.y)

        for (i, zpoint) in enumerate(Zpoints):
            Zpts = zpoint * np.ones(Xpts.shape)
            XY_data = self.ExtractDataFromDoseCube(Xpts, Ypts, Zpts)
            leg_strings.append(f'Z = {zpoint} mm')
            axs_ind = np.unravel_index(i, axs.shape)
            axs[axs_ind].imshow(XY_data, extent=extent)
            axs[axs_ind].set_xlabel('X [mm]', fontsize=FigureSpecs.LabelFontSize)
            axs[axs_ind].set_ylabel('Y [mm]', fontsize=FigureSpecs.LabelFontSize)
            axs[axs_ind].set_title(f'Dose at Z = {zpoint} mm [{self.dose.unit}]', fontsize=FigureSpecs.TitleFontSize)
            plt.tight_layout()
            plt.show()


def compare_multiple_results(BinFiles, abs_dose=False, custom_legend_names=None):
    """
    this produces depth dose and profile plots for a list of topas .bin files.

    :param BinFiles: list of files to analyse
    :type BinFiles: list
    :param abs_dose: % dose (False) or absolute dose (True)
    :type abs_dose: boolean, optional
    :param custom_legend_names: if passed, this list will be used for legend. If not, file names will be used.
    :type custom_legend_names: list, optional
    """

    LineStyles = ['C0-', 'C2--', 'C3:', 'C4-.', 'C5-', 'C6--', 'C7:']
    fig, axs = plt.subplots(ncols=2, nrows=1, figsize=[10, 5])
    legend_names = []
    for i, bin_file in enumerate(BinFiles):
        [path_name, file_name] = os.path.split(bin_file)
        legend_names.append(file_name)
        WTD = WaterTankData(path_name, file_name, AbsDepthDose=abs_dose)
        # plot the differences:

        PlotStyle = i % np.shape(LineStyles)[0]
        axs[0].plot(WTD.x, WTD.ProfileDose_X, LineStyles[PlotStyle], linewidth=2)
        axs[1].plot(WTD.z, np.flip(WTD.DepthDose), LineStyles[PlotStyle], linewidth=2)

        # sns.lineplot(ax=axs[0], x=WTD.x, y=WTD.ProfileDose_X)
        # sns.lineplot(ax=axs[1], x=WTD.z, y=np.flip(WTD.DepthDose))

    axs[0].set_xlabel('X [mm]', fontsize=FigureSpecs.LabelFontSize)
    axs[0].set_ylabel('Dose [%]', fontsize=FigureSpecs.LabelFontSize)
    axs[0].set_title('a) Dose profiles', fontsize = FigureSpecs.TitleFontSize)
    axs[0].grid()
    axs[1].set_xlabel('Z [mm]', fontsize=FigureSpecs.LabelFontSize)
    axs[1].set_ylabel('Dose [%]', fontsize=FigureSpecs.LabelFontSize)
    axs[1].set_title('b) Depth Dose', fontsize=FigureSpecs.TitleFontSize)
    axs[1].grid()

    if not custom_legend_names:
        axs[1].legend(legend_names, fontsize=FigureSpecs.LabelFontSize)
    else:
        axs[1].legend(custom_legend_names, fontsize=FigureSpecs.LabelFontSize)
    plt.show()


def ReadInLogFile(LogFileLoc):
    """
    Read in a log file and return as a dictionary

    :param LogFileLoc: path to log file
    :type LogFileLoc: string or pathlib.Path
    """

    if not os.path.isfile(LogFileLoc):
        print(f'File not found:\n{LogFileLoc}\nQuitting')
        sys.exit(1)

    file1 = open(LogFileLoc, 'r')
    Lines = file1.readlines()


    LineItteration = 0
    ResultsDict = {}
    for line in Lines:
        try:
            d = {i.split(': ')[0]: i.split(': ')[1] for i in line.split(', ')}
            # remaining keys are the things we want to track
            for key in d.keys():
                if key == 'Best parameter set' or 'length scales' in key:
                    # this is the last line
                    break
                if LineItteration == 0:
                    ResultsDict[key] = []
                ResultsDict[key].append(float(d[key]))
            LineItteration += 1
        except IndexError:
            pass

    return ResultsDict


def PlotLogFile(LogFileLoc, save_loc=None):
    """
    This function can be used to plot an existing log file

    :param LogFileLoc: location of log file
    :type LogFileLoc: string or pathlib.Path
    :param save_loc: if supplied, will attempt to save the log file there without displaying it.
        No error checking included! If not supplied, behaviour is instead to show the plot
    :type save_loc: string or pathlib.Path
    """
    ResultsDict = ReadInLogFile(LogFileLoc)
    Itteration = ResultsDict.pop('Itteration')
    OF = ResultsDict.pop('ObjectiveFunction')

    LowestVal = np.ones(np.size(OF)) * OF[0]
    for i, val in enumerate(LowestVal):
        if OF[i] < LowestVal[i]:
            LowestVal[i:] = OF[i]

    fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(10, 5))
    axs.plot(Itteration, LowestVal, '-k', linewidth=2)
    axs.plot(Itteration, OF, 'C6')
    axs.set_xlabel('Iteration number', fontsize=FigureSpecs.LabelFontSize)
    axs.set_ylabel('Objective function', fontsize=FigureSpecs.LabelFontSize)
    axs.grid(True)

    try:
        target_prediction = -1 * np.array(ResultsDict['target_prediction_mean'] )
        target_prediction_std = np.array(ResultsDict['target_prediction_std'])
        axs.plot(Itteration, target_prediction, 'C0')
        axs.fill_between(Itteration,
                         target_prediction + target_prediction_std,
                         target_prediction - target_prediction_std, alpha=0.15, color='C0')
        axs.legend(['Best', 'Actual', 'Predicted', r'$\sigma$'], fontsize=FigureSpecs.LabelFontSize)
    except KeyError:
        # predicted isn't  available for optimisers
        axs.legend(['Best', 'Current'])

    MinValue = np.argmin(OF)
    axs.plot(Itteration[MinValue], OF[MinValue], 'r-x')
    axs.set_title('Convergence Plot', fontsize=FigureSpecs.TitleFontSize)

    if save_loc:
        plt.savefig(save_loc)
        plt.close(fig)
    else:
        plt.show()




